'''Clip and simplify geometries'''
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, 2020, Neil Freeman <contact@fakeisthenewreal.org>
from functools import partial

try:
    from shapely.geometry import mapping, shape
    from shapely.geos import TopologicalError
except ImportError:
    pass
try:
    import numpy as np
except ImportError:
    pass
try:
    import visvalingamwyatt as vw
except ImportError:
    pass


def clipper(bbox):
    """
    Create a clipping function for a given bounding box.

    Args:
        bbox (tuple): bounding box

    Returns:
        function that will given geometries to input bounding box
    """
    minx, miny, maxx, maxy = bbox
    bounds = {
        "type": "Polygon",
        "coordinates": [[(minx, miny), (minx, maxy), (maxx, maxy), (maxx, miny), (minx, miny)]],
    }
    try:
        bbox_shape = shape(bounds)

        def func(geometry):
            # This is technically only needed in Py3, but whatever.
            try:
                clipped = bbox_shape.intersection(shape(geometry))
            except (ValueError, TopologicalError):
                return geometry

            return mapping(clipped)

    except NameError:

        def func(geometry):
            return geometry

    return func


def clip(geometry, bounds):
    """
    Clip a geometry to a bounding box. Equivalent to calling clipper(bounds)(geometry).

    Args:
        geometry (dict): geometry object
        bounds (tuple): bounding box

    Returns:
        (dict) geometry
    """
    try:
        return clipper(bounds)(geometry)

    except NameError:
        return geometry


def simplifier(ratio):
    """
    Create a simplification function, if visvalingamwyatt is available.
    Otherwise, return a noop function.

    Args:
        ratio (int): Between 1 and 99

    Returns:
        simplification function
    """
    try:
        # put this first to get NameError out of the way
        simplify = vw.simplify_geometry

        if ratio is None or ratio >= 100 or ratio < 1:
            raise SvgisError("Invalid ratio")

        return partial(simplify, ratio=ratio / 100.0)

    except (TypeError, ValueError, NameError):
        return None


def scale(coordinates, scalar=1):
    '''Scale a list of coordinates by a scalar. Only use with projected coordinates'''
    try:
        try:
            arr = np.array(coordinates, dtype=float)

        except TypeError:
            arr = np.array(list(coordinates), dtype=float)

        return arr * scalar

    except NameError:
        if isinstance(coordinates, tuple):
            return [coordinates[0] * scalar, coordinates[1] * scalar]

        return [(c[0] * scalar, c[1] * scalar) for c in coordinates]


def scale_rings(rings, factor=1):
    """Apply scale() to a list of rings."""
    return [scale(ring, factor) for ring in rings]


def scale_geom(geom, factor=1):
    """
    Scale a geometry by a given factor

    Args:
        geom (dict): geojson-like dict
        factor (numeric): scale factor, default: 1
    """
    if geom['type'] == 'MultiPolygon':
        geom['coordinates'] = [scale_rings(rings, factor) for rings in geom['coordinates']]

    elif geom['type'] in ('Polygon', 'MultiLineString'):
        geom['coordinates'] = scale_rings(geom['coordinates'], factor)

    elif geom['type'] in ('MultiPoint', 'LineString'):
        geom['coordinates'] = scale(geom['coordinates'], factor)

    elif geom['type'] == 'Point':
        geom['coordinates'] = scale(geom['coordinates'], factor)

    elif geom['type'] == 'GeometryCollection':
        geom['geometries'] = [scale_geom(i) for i in geom['geometries']]

    else:
        raise NotImplementedError(f"Unsupported geometry type: {geom['type']}")

    return geom



"""Tests on the svgis.map function"""
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>
# pylint: disable=duplicate-code
import os
import unittest
import warnings
from xml.dom import minidom

from svgis import svgis

warnings.filterwarnings("always")


class MapTestCase(unittest.TestCase):
    projection = '+proj=lcc +lat_1=20 +lat_2=60 +lat_0=40 +lon_0=-96 +x_0=0 +y_0=0 +datum=NAD83 +units=m +no_defs'
    bounds = (-124.0, 20.5, -64.0, 49.0)
    fixture = 'tests/fixtures/test.svg'
    shp = 'tests/fixtures/cb_2014_us_nation_20m.json'
    css = 'polygon{fill:green}'

    def testMapWithStyle(self):
        result = svgis.map(self.shp, style=self.css, scale=1000, crs=self.projection, bounds=self.bounds, clip=None)
        self.assertIn(self.css, result)

        style = 'tmp.css'

        with open(style, 'w') as w:
            w.write(self.css)

        try:
            result = svgis.map(self.shp, style=[style], scale=1000, crs=self.projection, bounds=self.bounds, clip=None)
            self.assertIn(self.css, result)

        finally:
            os.remove('tmp.css')

    def testMap(self):
        # pylint: disable=duplicate-code
        a = svgis.map(self.shp, scale=1000, crs=self.projection, bounds=self.bounds, clip=False)

        with open('a.svg', 'w') as A:
            A.write(a)

        try:
            result = minidom.parseString(a).getElementsByTagName('svg').item(0)
            fixture = minidom.parse(self.fixture).getElementsByTagName('svg').item(0)

            result_vb = [float(x) for x in result.attributes.get('viewBox').value.split(',')]
            fixture_vb = [float(x) for x in fixture.attributes.get('viewBox').value.split(',')]

            for r, f in zip(result_vb, fixture_vb):
                self.assertAlmostEqual(r, f, 5, 'viewbox doesnt match fixture')
        finally:
            os.remove("a.svg")

    def testMapProjFile(self):
        a = svgis.map(self.shp, scale=1000, crs='tests/fixtures/test.proj4', bounds=self.bounds, clip=False)

        result = minidom.parseString(a).getElementsByTagName('svg').item(0)
        fixture = minidom.parse(self.fixture).getElementsByTagName('svg').item(0)

        result_vb = [float(x) for x in result.attributes.get('viewBox').value.split(',')]
        fixture_vb = [float(x) for x in fixture.attributes.get('viewBox').value.split(',')]

        for r, f in zip(result_vb, fixture_vb):
            self.assertAlmostEqual(r, f, 5)


if __name__ == '__main__':
    unittest.main()



'''Create string versions of SVG elements.'''
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, 2020, Neil Freeman <contact@fakeisthenewreal.org>

from . import dom, utils


def _element(tag, contents=None, **kwargs):
    """
    Draw an element, optionally wrapping contents.

    Args:
        tag (str): tag name
        contents (str): contents to be wrapped in the tag
        kwargs: to be transformed into attributes

    Returns:
        ``str``
    """
    attribs = toattribs(**kwargs)
    if contents:
        return f"<{tag}{attribs}>{contents}</{tag}>"

    return f"<{tag}{attribs}/>"


def _fmt(precision):
    if precision is None:
        return "{0[0]},{0[1]}"
    return f"{{0[0]:.{precision}f}},{{0[1]:.{precision}f}}"


def _poly(name):
    def poly(coordinates, precision=None, **kwargs):
        fmt = _fmt(precision)

        points = utils.dedupe(fmt.format(c) for c in coordinates)
        return _element(name(), points=' '.join(points), **kwargs)

    return poly


def circle(point, precision=None, **kwargs):
    """
    Create a svg circle element. Keyword arguments are mapped to attributes.

    Args:
        point (tuple): The center of the circle
        precision (int): rounding precision

    Returns:
        ``str``
    """
    return _element('circle', cx=utils.rnd(point[0], precision), cy=utils.rnd(point[1], precision), **kwargs)


def text(string, start, precision=None, **kwargs):
    """
    Create an svg text element.

    Args:
        string (str): text for element
        start (tuple): starting coordinate

    Returns:
        ``str``
    """
    start = [utils.rnd(i, precision) for i in start]
    return _element('text', string, x=start[0], y=start[1], **kwargs)


def rect(start, width, height, precision=None, **kwargs):
    """
    Create an svg rect element.

    Args:
        start (tuple): starting coordinate
        width (int): rect width
        height (int): rect height
        precision (int): rounding precision

    Returns:
        ``str``
    """
    start = [utils.rnd(i, precision) for i in start]
    width = utils.rnd(width, precision)
    height = utils.rnd(height, precision)

    return _element('rect', x=start[0], y=start[1], width=width, height=height, **kwargs)


def line(start, end, precision=None, **kwargs):
    """
    Create an svg line element.

    Args:
        start (tuple): starting coordinate
        end (tuple): ending coordinate
        precision (int): rounding precision

    Returns:
        ``str``
    """
    start = [utils.rnd(i, precision) for i in start]
    end = [utils.rnd(i, precision) for i in end]

    return _element('line', x1=start[0], y1=start[1], x2=end[0], y2=end[1], **kwargs)


def path(coordinates, precision=None, **kwargs):
    """
    Create an svg path element as a string.

    Args:
        coordinates (Sequence): A sequence of coordinates and string instructions
        precision (int): rounding precision

    Returns:
        ``str``
    """
    fmt = _fmt(precision)
    coords = utils.dedupe(i if isinstance(i, str) else fmt.format(i) for i in coordinates)
    return _element('path', d=' '.join(coords), **kwargs)


@_poly
def polyline():
    """
    Create an svg polyline element

    Args:
        coordinates (Sequence): x, y coordinates
        precision (int): rounding precision

    Returns:
        ``str``
    """
    return 'polyline'


@_poly
def polygon():
    """
    Create an svg polygon element

    Args:
        coordinates (Sequence): x, y coordinates
        precision (int): rounding precision

    Returns:
        ``str``
    """
    return 'polygon'


def toattribs(**kwargs):
    """
    Convert keyword arguments to SVG attribute definitions.

    Returns:
        ``str``
    """
    attribs = ' '.join(f'{k}="{dom.ampencode(v)}"' for k, v in kwargs.items() if v is not None and v != '')
    if attribs:
        return ' ' + attribs

    return attribs


def defstyle(style=None):
    """
    Create a defs element that wraps a CSS style.

    Args:
        style (string): A CSS string.

    Returns:
        ``str``
    """
    if style:
        return f'<defs><style type="text/css"><![CDATA[{style}]]></style></defs>'

    return '<defs />'


def group(members=None, **kwargs):
    """
    Create a group with the given scale and translation.

    Args:
        members (Sequence): unicode SVG elements
        kwargs (dict): elements of this dictionary will be converted to
                        attributes of the group, i.e. key="value".

    Returns:
        ``str``
    """
    members = members or ''
    return _element('g', ''.join(members), **kwargs)


def drawing(size, members, precision=None, viewbox=None, style=None):
    """
    Create an SVG element.

    Args:
        size (tuple): width, height
        members (list): Strings to add to output.
        viewbox (Sequence): Four coordinates that describe an SVG viewBox.
        style (string): CSS string.

    Returns:
        ``str``
    """
    kwargs = {
        'width': size[0],
        'height': size[1],
        'baseProfile': 'full',
        'version': '1.1',
        'xmlns': 'http://www.w3.org/2000/svg',
    }

    if precision:
        fmt = ('{:.{precision}f}',)
    else:
        fmt = ('{:f}',)

    if viewbox:
        kwargs['viewBox'] = (','.join(fmt * 4)).format(*viewbox, precision=precision)

    contents = defstyle(style) + ''.join(members)
    return _element('svg', contents, **kwargs)



# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://www.opensource.org/licenses/GNU General Public License v3 (GPLv3)-license
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>

# pylint: disable=missing-module-docstring,missing-function-docstring,missing-class-docstring

import logging

logger = logging.getLogger('svgis')
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)

TEST_SVG = """<svg baseProfile="full" height="1" version="1.1" xmlns="http://www.w3.org/2000/svg">
    <defs></defs>
    <g>
        <g id="test">
            <polygon class="test" points="3,2 -2,6 8,-1 8,2 4,1 3,2" />
        </g>
        <g id="foo">
            <polyline id="baz" class="foo" points="3,2 -2,6 8,-1"></polyline>
        </g>
        <g id="cat">
            <polyline id="meow" class="long-class-name" points="3,2 -2,6 8,-1"></polyline>
            <circle id="orb" cx="1" cy="1"></circle>
        </g>
    </g>
</svg>"""

TEST_CSS = """
polygon {fill: orange;}
.test { stroke: green; }
polyline { stroke: blue}
.test, #baz { stroke-width: 2; }
#test ~ #foo { fill: purple; }
#cat polyline { fill: red }
#meow { stroke-opacity: 0.50 }
circle { r: 1; }
"""



# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>
import unittest

from svgis import graticule
from svgis.errors import SvgisError


class GraticuleTestCase(unittest.TestCase):
    def testCRS(self):
        g = graticule.graticule((16.34, -34.81, 32.83, -22.09), step=10000, crs_or_method='utm')
        a = next(g)
        self.assertIsInstance(a, dict)

    def testErr(self):
        with self.assertRaises(SvgisError):
            g = graticule.graticule((16.34, -34.81, 32.83, -22.09), step=10000, crs_or_method='file')
            next(g)

    def test_feature(self):
        g = graticule._feature(0, [1, 2, 3])
        self.assertEqual(g['geometry'], {'type': 'LineString', 'coordinates': [1, 2, 3]})
        self.assertEqual(g['type'], 'Feature')
        self.assertEqual(g['id'], 0)

    def test_layer(self):
        a = graticule.layer([0, 0, 5, 5], 1)
        assert isinstance(a, dict)
        assert isinstance(a['features'], (list, tuple))
        self.assertIsInstance(a['features'][0]['geometry']['coordinates'], (list, tuple))

    def testgraticule(self):
        g = graticule.graticule((0, 0, 2, 2), 1)

        fixture1 = [(0, i / 2.0) for i in range(9)]
        fixture2 = [(1, i / 2.0) for i in range(9)]

        self.assertSequenceEqual(list(next(g).get('geometry').get('coordinates')), fixture1)
        self.assertSequenceEqual(list(next(g).get('geometry').get('coordinates')), fixture2)


if __name__ == '__main__':
    unittest.main()



'''Utilities for working with bounding boxes'''
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, 2020, Neil Freeman <contact@fakeisthenewreal.org>
from pyproj.transformer import Transformer

from . import errors, utils


def check(bounds):
    '''Check if bounds are valid.'''
    # Refuse to set these more than once
    try:
        if bounds is None or len(bounds) != 4 or not all(bounds):
            return False

    except (TypeError, AttributeError, ValueError):
        return False

    if bounds[0] > bounds[2]:
        bounds = bounds[2], bounds[1], bounds[0], bounds[3]

    if bounds[1] > bounds[3]:
        bounds = bounds[0], bounds[3], bounds[2], bounds[1]

    return bounds


def update(old, new):
    """
    Extend old with any more distant values from newpoints.
    Also replace any missing min/max points in old with values from new.
    """
    bounds = []
    inf = float('inf')
    neginf = inf * -1

    # python3 gives TypeError when using None in min/max
    # This contraption avoids that problem.
    # List comp below replaces Nones in bounds with real values in new or old
    for n, m in zip(new[:2], old[:2]):
        try:
            if neginf in (m, n):
                bounds.append(max(n, m))
                continue

            bounds.append(min(n, m))
        except TypeError:
            bounds.append(None)

    for n, m in zip(new[2:], old[2:]):
        try:
            if inf in (m, n):
                bounds.append(min(n, m))
                continue

            bounds.append(max(n, m))
        except TypeError:
            bounds.append(None)

    if any(not v for v in bounds):
        bounds = list((a or b or c) for a, b, c in zip(bounds, new, old))

    return bounds


def pad(bounds, ext=100):
    """
    Pad a bounding box. Works best when input is in a projected unit (i.e. feet or meters).

    Args:
        bounds (tuple): a bounding box (minx, miny, maxx, maxy)
        ext (int): the distance to pad the box, in projected units. Default: 100.

    Returns:
        tuple: A bounding box (minx, miny, maxx, maxy)
    """
    try:
        return bounds[0] - ext, bounds[1] - ext, bounds[2] + ext, bounds[3] + ext
    except TypeError:
        return bounds


def ring(bounds):
    '''
    Convert min, max points to a boundary ring.

    Args:
        bounds (tuple): a bounding box (minx, miny, maxx, maxy)

    Return:
        list: a geojson-like ring of coordinates

    '''
    minx, miny, maxx, maxy = bounds
    xs, ys = list(utils.between(minx, maxx)), list(utils.between(miny, maxy))

    left_top = [(minx, y) for y in ys] + [(x, maxy) for x in xs][1:]

    ys.reverse()
    xs.reverse()

    return left_top + [(maxx, y) for y in ys] + [(x, miny) for x in xs]


def covers(b1, b2):
    """Check if a bounding box covers another.

    Args:
        b1 (tuple): a bounding box (minx, miny, maxx, maxy)
        b2 (tuple): a bounding box

    Returns:
        bool: ``False`` if any points in ``b2`` are outside ``b1``.
    """
    return b1[0] <= b2[0] and b1[1] <= b2[1] and b1[2] >= b2[2] and b1[3] >= b2[3]


def transform(bounds, **kwargs):
    """
    Project a bounding box, taking care to not slice off the sides.

    Args:
        bounds (tuple): bounding box to transform.
        transformer (pyproj.transformer.Transformer): A pyproj Transformer instance.
        in_crs (dict): Fiona-type proj4 mapping representing input projection.
        out_crs (dict): Fiona-type proj4 mapping representing output projection.

    Returns:
        tuple: The input ``bounds`` reprojected from ``in_crs`` to ``out_crs``
    """
    transformer = kwargs.get('transformer')
    in_crs = kwargs.get('in_crs')
    out_crs = kwargs.get('out_crs')

    if not transformer and not (in_crs and out_crs):
        raise errors.SvgisError('Need input CRS and output CRS or a Transformer')

    if transformer is None:
        transformer = Transformer.from_crs(in_crs, out_crs, always_xy=True)

    densebounds = ring(bounds)
    xbounds, ybounds = list(zip(*transformer.itransform(densebounds)))
    return min(xbounds), min(ybounds), max(xbounds), max(ybounds)



'''Mess around with SVG styling'''
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, 2020, Neil Freeman <contact@fakeisthenewreal.org>
# pylint: disable=c-extension-no-member
import logging
import os.path
import re
from string import ascii_letters

import tinycss2
from lxml import etree

from . import dom

LOG = logging.getLogger('svgis')


def sanitize(string):
    """
    Make input safe of use in an svg ID or class field.
    Replaces blocks of whitespace with an underscore (``_``),
    deletes periods, number signs and double-quotes (``.#"``).
    If the first character isn't an ascii letter, dash (``-``)
    or underscore (``_``), an underscore is added to the beginning.

    Args:
        string (mixed): Input to sanitize

    Returns:
        str
    """
    try:
        string = re.sub(r'\s+', '_', re.sub(r'(\.|#|"|&)', '', string))
        return string if string[:1] in '_-' + ascii_letters else '_' + string

    except TypeError:
        return sanitize(str(string))


def construct_classes(classes, properties):
    """
    Build a class string for an element using the properties. Class names
    take the form ``myclass_value``. If a given class isn't found in properties,
    the class name is added (e.g. ``myclass``).

    Args:
        classes (Sequence): Column names to include in the class list
        properties (dict): A single feature's properties.

    Returns:
        (list) class names
    """
    f = '{}_{}'
    return [sanitize(f.format(p, properties.get(p))) for p in classes if p in properties]


def construct_datas(fields, properties):
    """
    Build a data- attribute string for an element using the properties. Attributes
    take the form data_FIELD=PROPERTY.

    Args:
        datas (Sequence): Column names to include in the class list
        properties (dict): A single feature's properties.

    Returns:
        (dict) attribute dictionary
    """
    return {sanitize('data-' + n): str(properties.get(n)) for n in fields if n in properties}


def pick(style):
    """
    Fetch a CSS string.

    Args:
        style (str): Either a CSS string or the path of a css file.
    """
    try:
        _, ext = os.path.splitext(style)
        if ext == '.css':
            with open(style, encoding="utf-8") as f:
                return f.read()

    except (AttributeError, TypeError):
        # Probably style is None.
        return None

    except IOError:
        logging.getLogger('svgis').warning("Couldn't read %s, proceeding with default style", style)

    return style


def rescale(svgfile, factor):
    """Add a ``scale()`` operation to an entire svg file."""
    svg = etree.parse(svgfile).getroot()
    scalar = f'scale({factor})'
    g = svg.find('.//g', namespaces=svg.nsmap)
    g.attrib['transform'] = (g.attrib.get('transform') + ' ' + scalar).strip()
    return etree.tostring(svg, encoding='utf-8').decode('utf-8')


def replace_comments(css):
    """
    Replace one-line non-standard comments with body-style comments::

       // non-standard comment
       /* body-style comment */

    """
    return re.sub(r'//(.+)', r'/*\1 */', css)


def add_style(svgfile, style, replace=False):
    """
    Add to or replace the CSS style in an SVG file.

    Args:
        svgfile (str): Path to an SVG file or an SVG string.
        newstyle (str): CSS string, or path to CSS file.
        replace (bool): If true, replace the existing CSS with newstyle (default: False)
    """
    if style == '-':
        style = '/dev/stdin'

    root, ext = os.path.splitext(style)

    if ext == '.css' or root == '/dev/stdin':
        with open(style, encoding="utf-8") as f:
            style = replace_comments(f.read())

    try:
        svg = etree.parse(svgfile).getroot()
    except IOError:
        try:
            svg = etree.fromstring(svgfile)
        except UnicodeDecodeError:
            svg = etree.fromstring(svgfile.encode('utf-8'))

    defs = svg.find('defs', namespaces=svg.nsmap)

    if defs is None:
        defs = etree.Element('defs', nsmap=svg.nsmap)
        svg.insert(0, defs)

    style_element = defs.find('.//style', namespaces=svg.nsmap)

    if style_element is None:
        style_element = etree.Element('style', nsmap=svg.nsmap)
        defs.append(style_element)

    if replace:
        style_content = style
    else:
        # Append CSS.
        style_content = (style_element.text or '') + ' ' + style

    # append cdata
    style_element.text = etree.CDATA(style_content)

    return etree.tostring(svg, encoding='utf-8').decode('utf-8')


def inline(svg):
    """
    Inline the CSS rules in an SVG. This is a very rough operation,
    and full css precedence rules won't be respected. May ignore sibling
    operators (``~``, ``+``), pseudo-selectors (e.g. ``:first-child``), and
    attribute selectors (e.g. ``.foo[name=bar]``). Works best with rules like:

    * ``.class``
    * ``tag``
    * ``tag.class``
    * ``#layer .class``
    * ``#layer tag``

    Args:
        svg (string): An SVG document.
        style (string): CSS to use, instead of the CSS in the <defs> element of the SVG.
    """
    try:
        doc = etree.fromstring(svg.encode('utf-8'))
        style_element = doc.find('.//style', namespaces=doc.nsmap)
        if style_element is None:
            return svg

        css = style_element.text
        rules = tinycss2.parse_stylesheet(css, skip_whitespace=True, skip_comments=True)
        dom.apply_rules(doc, rules)
        return etree.tostring(doc, encoding='utf-8').decode('utf-8')

    # Return plain old SVG.
    except (AttributeError, NameError) as e:
        logging.getLogger('svgis').warning("Unable to inline CSS: %s", e)
        return svg



"""Do the work of picking, generating and transforming coordinate reference systems."""
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, 2020, Neil Freeman <contact@fakeisthenewreal.org>
import logging
import os.path

import utm
from pyproj.crs import CRS
from pyproj.exceptions import CRSError

from . import bounding, errors
from .utils import DEFAULT_GEOID

LOG = logging.getLogger('svgis')
METHODS = 'default', 'file', 'local', 'utm'


def tm_proj4(x0, y0, y1):
    """
    Generate the proj4 string for a local Transverse Mercator projection
    centered at a given longitude and between two latitudes.

    Args:
        x0 (float): longitude
        y0 (float): latitude 0
        y1 (float): latitude 1

    Returns:
        (str) proj4 string
    """
    proj = f'+proj=lcc +lon_0={x0} +lat_1={y1} +lat_2={y0} +lat_0={y1}'
    return proj + ' +x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0 +units=m +no_defs'


def utm_proj4(lon, lat):
    """
    Generate the proj4 string for the UTM projection at a given (lon, lat) coordinate.

    Args:
        lon (float): longitude
        lat (float): latitude

    Returns:
        (str) proj4 string
    """
    try:
        _, _, zonenumber, zoneletter = utm.from_latlon(lat, lon)

        if zoneletter in 'ZYXWVUTSRQPN':
            hemisphere = 'north'

        elif zoneletter in 'MLKJHGFEDCBA':
            hemisphere = 'south'

        return f'+proj=utm +zone={zonenumber} +{hemisphere} +datum=WGS84 +units=m +no_defs'

    except utm.error.OutOfRangeError as err:
        raise errors.SvgisError(err) from err


def generateproj4(method, bounds, file_crs):
    """
    Generate a Proj4 projection definition: either the local UTM zone
    or a custom transverse mercator.

    Args:
        method (str): If default, local or utm.
            * If 'utm': generate a UTM. Otherwise, a local projection,
            * If 'local': generate a custom local transverse mercator projection,
            * If 'default': if file_crs is longlat, act like local. otherwise use file_crs
        bounds (tuple): bounding box
        file_crs (dict): Fiona-generated CRS of the input file

    Returns:
        (str) proj4 string
    """
    LOG.debug('generating proj4')
    if bounds is None or file_crs is None:
        raise errors.SvgisError('generatecrs missing bounds and file crs')

    is_longlat = _is_longlat(file_crs)
    if method == 'default':
        # Check if file_crs _is_longlat, if not use that.
        # Option to continue returning default if we didn't get a file_crs
        if is_longlat:
            method = 'local'
        else:
            return CRS(file_crs)

    if is_longlat:
        longlat_bounds = bounds
    else:
        longlat_bounds = bounding.transform(bounds, in_crs=file_crs, out_crs=DEFAULT_GEOID)

    minx, miny, maxx, maxy = longlat_bounds

    if method == 'utm':
        midx = (minx + maxx) / 2
        midy = (miny + maxy) / 2
        return utm_proj4(midx, midy)

    if method == 'local':
        # Create a custom TM projection
        x0 = (float(minx) + float(maxx)) // 2

        return tm_proj4(x0, miny, maxy)

    raise errors.SvgisError(f"Unexpected method. Valid methods are default, local or utm. Got: {method}")


def _is_longlat(crs):
    '''Test if CRS is in lat/long coordinates'''
    try:
        return crs['proj'] == 'longlat'
    except (KeyError, TypeError, AttributeError):
        pass

    projection = CRS(crs)

    try:
        return projection.is_geographic
    except AttributeError:
        return crs.to_dict().get('proj') == 'longlat'


def pick(project, bounds=None, file_crs=None):
    """
    Pick a projection or projection method to use.

    Returns:
        (mixed) one of: None, 'local', 'utm' or a dict
    """
    LOG.debug('projection.pick("%s")', project)
    project = project or 'default'
    if isinstance(project, CRS):
        return project

    try:
        return CRS(project)
    except CRSError:
        pass

    if isinstance(project, str):
        LOG.debug('projection is a str')
        if project.lower() == 'file':
            LOG.debug('"project" == file')
            return file_crs if file_crs is not None else 'file'

        if project.lower() in METHODS:
            LOG.debug('"project" is in METHODS')
            return CRS(generateproj4(project, bounds, file_crs))

        if os.path.exists(project):
            with open(project) as f:
                return CRS(f.read())

    raise errors.SvgisError(f'Unable to convert to projection: {project}')


def fake_to_string(crs):
    """
    Fake to_string for debugging in places where fiona.crs.to_string
    isn't available
    """
    return ' '.join(f'+{i[0]}={i[1]}' for i in crs.items())



"""Create SVG drawings from vector geodata files (SHP, geoJSON, etc)."""
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>
# pylint: disable=redefined-builtin
from . import bounding, draw, errors, projection, style, svg, svgis, transform
from .svgis import SVGIS, map

__version__ = '0.5.3'

__all__ = [
    'bounding',
    'draw',
    'errors',
    'projection',
    'style',
    'svg',
    'svgis',
    'transform',
]



"""Tests on bounding box management."""
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>
import unittest

from svgis import bounding, errors


class ConvertTestCase(unittest.TestCase):
    def test_updatebounds(self):
        bounds1 = (None, 0.1, None, 1.1)
        bounds2 = (0.2, 0.2, 1.2, 1.2)
        bounds3 = (0.1, 0.3, 1.5, 1.1)
        bounds4 = (0.05, 0.4, float('inf'), 1.2)
        bounds5 = (0.05, -1 * float('inf'), 1.4, 1.2)

        self.assertSequenceEqual(bounding.update(bounds1, bounds2), (0.2, 0.1, 1.2, 1.2))
        self.assertSequenceEqual(bounding.update(bounds3, bounds2), (0.1, 0.2, 1.5, 1.2))
        self.assertSequenceEqual(bounding.update(bounds3, bounds4), (0.05, 0.3, 1.5, 1.2))
        self.assertSequenceEqual(bounding.update(bounds3, bounds5), (0.05, 0.3, 1.5, 1.2))

    def testConvertBbox(self):
        bounds = (-100, -100, 100, 100)

        self.assertSequenceEqual(bounding.pad(bounds, ext=100), (-200, -200, 200, 200))
        self.assertSequenceEqual(bounding.pad(bounds, ext=10), (-110, -110, 110, 110))

    def test_bbox_covers(self):
        a = (0, 0, 10, 10)
        b = (0, 0, 20, 10)
        c = (0, 0, 5, 11)

        self.assertFalse(bounding.covers(a, b))
        self.assertTrue(bounding.covers(b, a))
        self.assertFalse(bounding.covers(a, c))
        self.assertFalse(bounding.covers(c, a))
        self.assertTrue(bounding.covers(c, c))

    def testbounds_to_ring(self):
        fixture = [
            (0, 0),
            (0, 0.6),
            (0, 1.2),
            (0, 1.7999999999999998),
            (0, 2.4),
            (0, 3.0),
            (0, 3.6),
            (0, 4.2),
            (0, 4.8),
            (0, 5.3999999999999995),
            (0, 5.999999999999999),
            (0.6, 6),
            (1.2, 6),
            (1.7999999999999998, 6),
            (2.4, 6),
            (3.0, 6),
            (3.6, 6),
            (4.2, 6),
            (4.8, 6),
            (5.3999999999999995, 6),
            (5.999999999999999, 6),
            (6, 5.999999999999999),
            (6, 5.3999999999999995),
            (6, 4.8),
            (6, 4.2),
            (6, 3.6),
            (6, 3.0),
            (6, 2.4),
            (6, 1.7999999999999998),
            (6, 1.2),
            (6, 0.6),
            (6, 0),
            (5.999999999999999, 0),
            (5.3999999999999995, 0),
            (4.8, 0),
            (4.2, 0),
            (3.6, 0),
            (3.0, 0),
            (2.4, 0),
            (1.7999999999999998, 0),
            (1.2, 0),
            (0.6, 0),
            (0, 0),
        ]

        r = bounding.ring((0, 0, 6, 6))
        self.assertSequenceEqual(r, fixture)

    def testTransformBounds(self):
        bounds = (-74, 42, -73, 43)

        with self.assertRaises(errors.SvgisError):
            bounding.transform(bounds, in_crs=4269)

        with self.assertRaises(errors.SvgisError):
            bounding.transform(bounds, out_crs=4269)

        a = bounding.transform(bounds, in_crs=4269, out_crs=3102)

        fixture = (43332271.446783714, 15585187.3924282, 44004528.34377961, 16321716.827537995)
        for z in zip(a, fixture):
            self.assertAlmostEqual(*z, places=5)

    def testCheck(self):
        fixture = (1, 1, 2, 2)
        result = bounding.check((2, 2, 1, 1))
        self.assertSequenceEqual(result, fixture)

        result = bounding.check((2, 1, 1, 2))
        self.assertSequenceEqual(result, fixture)

        result = bounding.check((1, 2, 2, 1))
        self.assertSequenceEqual(result, fixture)

        self.assertFalse(bounding.check(None))
        self.assertFalse(bounding.check(object))
        self.assertFalse(bounding.check((1, 2, 3)))
        self.assertFalse(bounding.check((1, 2, 3, None)))


if __name__ == '__main__':
    unittest.main()



# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>
import unittest

from svgis import utils
from svgis.errors import SvgisError


class UtilsTestCase(unittest.TestCase):
    def test_isinf(self):
        self.assertTrue(utils.isinf(float('inf')))
        self.assertTrue(utils.isinf(float('-inf')))
        self.assertFalse(utils.isinf(float(10e10)))

    def test_between(self):
        self.assertSequenceEqual(list(utils.between(0.0, 10.0, 10)), [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0])

        self.assertSequenceEqual(list(utils.between(0.0, 5.0)), [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5])

    def test_frange(self):
        self.assertSequenceEqual(list(utils.frange(1.0, 2.15, 0.25)), [1.0, 1.25, 1.5, 1.75, 2.0])
        self.assertSequenceEqual(list(utils.frange(1.0, 2.15, 0.25, True)), [1.0, 1.25, 1.5, 1.75, 2.0, 2.25])

    def test_modfloor(self):
        self.assertEqual(utils.modfloor(10, 3), 9)
        self.assertEqual(utils.modfloor(17, 4), 16)

    def test_modceil(self):
        self.assertEqual(utils.modceil(10, 3), 12)
        self.assertEqual(utils.modceil(17, 4), 20)

    def test_dedupe(self):
        test = [1, 2, 2, 3, 4, 5, 4]
        fixture = [1, 2, 3, 4, 5, 4]
        self.assertSequenceEqual(list(utils.dedupe(test)), fixture)

        test = [(1, 2), (3, 4), (3, 4), (5, 4), 'M', (3, 4)]
        fixture = [(1, 2), (3, 4), (5, 4), 'M', (3, 4)]

        self.assertSequenceEqual(list(utils.dedupe(test)), fixture)


if __name__ == '__main__':
    unittest.main()



"""Tests on the map styling tools, most of which are CSS manipulation."""
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>
# pylint: disable=unused-import
import os
import re
import unittest
from io import BytesIO, StringIO
from xml.dom import minidom

try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree

from svgis import dom, style

from . import TEST_CSS, TEST_SVG


class CssTestCase(unittest.TestCase):
    svg = TEST_SVG
    css = TEST_CSS

    css1 = '''.class-name { fill: orange;}'''

    file = 'tests/fixtures/test.svg'

    classes = ('apple', 'potato')

    properties = {
        'apple': 'fruit',
        'pear': 1,
        'kale': 'leafy green',
    }

    def testInlineCSS(self):
        svg = style.add_style(self.svg, self.css)
        inlined = style.inline(svg)
        self.assertNotEqual(inlined, self.svg)

        doc = minidom.parseString(inlined)
        self.assertIn('fill:purple', inlined)

        polygon_style = doc.getElementsByTagName('polygon').item(0).getAttribute('style')
        self.assertIn('stroke:green', polygon_style)
        self.assertIn('fill:orange', polygon_style)

        cat_style = doc.getElementsByTagName('polyline').item(1).getAttribute('style')
        self.assertIn('fill:red', cat_style)

        polyline_style = doc.getElementsByTagName('polyline').item(0).getAttribute('style')
        self.assertIn('stroke:blue', polyline_style)

    def test_add_style(self):
        new = style.add_style(self.file, self.css)
        result = minidom.parseString(new).getElementsByTagName('defs').item(0).getElementsByTagName('style').item(0)
        assert self.css in result.toxml()

    def test_replace_style(self):
        new = style.add_style(self.file, self.css, replace=True)
        result = minidom.parseString(new).getElementsByTagName('defs').item(0).getElementsByTagName('style').item(0)
        assert self.css in result.toxml()

    def test_add_style_missing_def(self):
        with open(self.file) as f:
            replaced_svg = re.sub(r'<defs></defs>', '', f.read())

        try:
            io_svg = BytesIO(replaced_svg)
        except TypeError:
            io_svg = StringIO(replaced_svg)

        new = style.add_style(io_svg, self.css)
        result = minidom.parseString(new).getElementsByTagName('defs').item(0).getElementsByTagName('style').item(0)
        self.assertIn(self.css, result.toxml())

    def testReScale(self):
        result = style.rescale('tests/fixtures/test.svg', 1.37)
        self.assertIn('scale(1.37)', result[0:2000])

    def testPickStyle(self):
        stylefile = 'tmp.css'

        with open(stylefile, 'w') as w:
            w.write(self.css)

        try:
            result = style.pick(stylefile)
            self.assertEqual(self.css, result)

        finally:
            os.remove('tmp.css')

        result = style.pick(self.css)
        self.assertEqual(self.css, result)

        self.assertIsNone(style.pick(None))

    def testAddCli(self):
        result = style.add_style(self.file, self.css)
        self.assertIn(self.css, result[0:2000])

        cssfile = 'tmp.css'
        with open(cssfile, 'w') as w:
            w.write(self.css)

        try:
            result = style.add_style(self.file, cssfile)
            self.assertIn(self.css, result[0:2000])

        finally:
            os.remove('tmp.css')

    def testAddStyleNoDefs(self):
        svg = self.svg.replace('<defs></defs>', '')
        new = style.add_style(svg, self.css)
        result = minidom.parseString(new).getElementsByTagName('defs').item(0).getElementsByTagName('style').item(0)
        assert self.css in result.toxml()

    def testSanitize(self):
        assert style.sanitize(None) == 'None'
        assert style.sanitize('') == ''
        self.assertEqual(style.sanitize('ü'), '_ü')
        self.assertEqual(style.sanitize('!foo'), '_!foo')
        assert style.sanitize('müller') == 'müller'

        self.assertEqual(style.sanitize(1), '_1')

        self.assertEqual(style.sanitize('foo.bar'), 'foobar')
        self.assertEqual(style.sanitize('fooba.r'), 'foobar')

        self.assertEqual(style.sanitize('.foo'), 'foo')

        self.assertEqual(style.sanitize('foo#bar'), 'foobar')
        self.assertEqual(style.sanitize('foobar#'), 'foobar')

        self.assertEqual(style.sanitize('x \t'), 'x_')

        self.assertEqual(style.sanitize('"huh"'), 'huh')

    def testConstructClasses(self):
        self.assertEqual(style.construct_classes(('foo',), {'foo': 'bar'}), ['foo_bar'])
        self.assertEqual(style.construct_classes(['foo'], {'foo': 'bar'}), ['foo_bar'])

        self.assertEqual(style.construct_classes(['foo'], {'foo': None}), ['foo_None'])

    def testCreateClasses(self):
        classes = style.construct_classes(self.classes, self.properties)
        self.assertEqual(classes, ['apple_fruit'])

        classes = style.construct_classes(self.classes, {'apple': 'fruit'})
        self.assertEqual(classes, ['apple_fruit'])

        classes = style.construct_classes(self.classes, {'apple': 'früit'})
        self.assertEqual(classes, ['apple_früit'])

        classes = style.construct_classes(self.classes, {'apple': 1})
        self.assertEqual(classes, ['apple_1'])

    def testCreateClassesMissing(self):
        classes = style.construct_classes(self.classes, {'apple': ''})
        self.assertEqual(classes, ['apple_'])

        classes = style.construct_classes(self.classes, {'apple': None})
        self.assertEqual(classes, ['apple_None'])

    def testPartialStyleName(self):
        inlined = style.inline(self.svg)
        self.assertNotIn('orange', inlined)

    def testReplaceComments(self):
        css = """
        // foo
        """
        result = style.replace_comments(css)
        self.assertIn('/* foo */', result)

    def testConstructDataFields(self):
        fields = style.construct_datas(['a', 'b'], {'a': '1', 'b': '_ _'})
        self.assertEqual(fields['data-a'], '1')
        self.assertEqual(fields['data-b'], '_ _')
        self.assertNotIn('data-c', fields)


if __name__ == '__main__':
    unittest.main()



'''Utilities for manipulating the DOM and applying styles to same'''
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, 2020, Neil Freeman <contact@fakeisthenewreal.org>
# pylint: disable=c-extension-no-member
import logging
import re

from lxml.cssselect import CSSSelector

SVG_NS = 'http://www.w3.org/2000/svg'
LOG = logging.getLogger('svgis')


def ampencode(value):
    """Escape an ampersand that isn't already url encoded"""
    return re.sub(r'&(?!amp;)', r'&amp;', str(value))


def serialize_token(token, previous_token=None):
    """
    Convert a tinycss2 selector to string in preparation for use with cssselect.
    This involves prepending 'svg|'  to element selectors.

    Arguments:
        token (tinycss2.ast.Node): The selector to serialize.
    """
    prev_type = 'whitespace' if previous_token is None else previous_token.type
    if token.type == 'ident' and prev_type == 'whitespace':
        return 'svg|' + token.serialize()
    return token.serialize()


def serialize_prelude(rule):
    """Convert the selector section of a CSS rule to string."""
    p = (serialize_token(a, b) for a, b in zip(rule.prelude, [None] + rule.prelude[:-1]))
    return ''.join(p).strip()


def rule_content(rule):
    """Except the content from CSS rules."""
    return [token.serialize() for token in rule.content if token.type != 'whitespace']


def apply_rules(doc, rules, nsmap=None):
    """
    Apply tinycss2 rules to an etree.Element (only tested on documents created by SVGIS).

    Args:
        doc (ElementTree.Element): The svg document to scan.
        rules (list): List of tinycss2 Rules to apply.
        nsmap (dict): namespace map.  Default: ``{ "svg": "http://www.w3.org/2000/svg" }``
    """
    nsmap = nsmap or {'svg': SVG_NS}
    for rule in rules:
        apply_rule(doc, rule, nsmap)

    return doc


def apply_rule(doc, rule, nsmap):
    """
    Apply a tinycss2 rule to an etree.Element
    Args:
        doc (ElementTree.Element): The svg document to scan.
        rule (QualifiedRule): tinycss2 rule
        nsmap (dict): namespace map
    """
    declaration = serialize_prelude(rule)
    selector = CSSSelector(declaration, namespaces=nsmap)

    tokens = rule_content(rule)
    rvalue = None
    if 'r' in tokens:
        i = tokens.index('r')
        token = rule.content[i]
        for token in rule.content[i + 1 :]:
            if token.type == 'number':
                rvalue = token.serialize()
                break

    rule_attrib = ''.join(tokens)
    LOG.debug('rule :%s', rule)

    for el in selector(doc):
        style_attrib = el.attrib.get('style', '') + ';'
        el.attrib['style'] = (style_attrib + rule_attrib).strip('; ')
        if rvalue:
            el.attrib['r'] = rvalue



'''Draw regular-interval graticules for a coordinate system'''
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://www.opensource.org/licenses/GNU General Public License v3 (GPLv3)-license
# Copyright (c) 2016, 2020, Neil Freeman <contact@fakeisthenewreal.org>
import json
from functools import partial

from pyproj.transformer import Transformer

from . import bounding, errors, projection, utils


def graticule(bounds, step, crs_or_method=None):
    """
    Draw graticules.

    Args:
        bounds (tuple): In WGS84 coordinates.
        step (int): Distance between graticule lines, in the output projection.
        crs_or_method (str): A projection specification.

    Returns:
        A generator that yields GeoJSON-like dicts of graticule features.
    """
    if crs_or_method == 'file':
        raise errors.SvgisError("'file' is not a valid option for projecting graticules.")

    if crs_or_method:
        out_crs = projection.pick(crs_or_method, bounds=bounds, file_crs=utils.DEFAULT_GEOID)
        unproject = Transformer.from_crs(utils.DEFAULT_GEOID, out_crs, skip_equivalent=True, always_xy=True)
        bounds = bounding.transform(bounds, transformer=unproject)

    else:
        unproject = Transformer.from_crs(4269, 4269, always_xy=True, skip_equivalent=True)

    minx, miny, maxx, maxy = bounds

    minx, miny = utils.modfloor(minx, step), utils.modfloor(miny, step)
    maxx, maxy = utils.modceil(maxx, step), utils.modceil(maxy, step)

    frange = partial(utils.frange, cover=True)

    i = 0
    for i, X in enumerate(frange(minx, maxx + step, step), 1):
        coords = unproject.itransform([(X, y) for y in frange(miny, maxy + step, step / 2.0)])
        yield _feature(i, coords, axis='x', coord=X)

    for i, Y in enumerate(frange(miny, maxy + step, step), i + 1):
        coords = unproject.itransform([(x, Y) for x in frange(minx, maxx + step, step / 2.0)])
        yield _feature(i, coords, axis='y', coord=Y)


def _feature(i, coords, axis=None, coord=None):
    return {
        'geometry': {'type': 'LineString', 'coordinates': list(coords)},
        'properties': {'axis': axis, 'coord': coord},
        'type': 'Feature',
        'id': i,
    }


def layer(bounds, step, crs=None):
    """
    Returns a graticule FeatureCollection object over the given bounds.

    Args:
        bounds (tuple): minx, miny, maxx, maxy
        step (numeric): distance between graticule lines
        crs: a CRS object or SRID.
    Returns:
        dict
    """
    return {
        "type": "FeatureCollection",
        "crs": {
            "type": "name",
            "properties": {"name": "urn:ogc:def:crs:OGC:1.3:CRS84"},
        },
        "features": tuple(graticule(bounds, step, crs)),
    }


def geojson(bounds, step, crs=None):
    """
    Create a geojson with graticules of the given bounds.

    Returns:
        str
    """
    return json.dumps(layer(bounds, step, crs))



# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>
import unittest
from xml.dom import minidom

import six

from svgis import style, svg


class SvgTestCase(unittest.TestCase):
    def setUp(self):
        self.file = 'tests/fixtures/test.svg'
        self.newstyle = 'stroke {color:red;}'

    def test_rescale(self):
        new = style.rescale(self.file, 100)

        g = minidom.parseString(new).getElementsByTagName('g')[0]
        assert 'scale(100)' in g.attributes.get('transform').value

    def test_create(self):
        s = svg.drawing((100, 100), [])
        self.assertIsInstance(s, six.string_types)

        s = svg.drawing((100, 100), [])
        self.assertIsInstance(s, six.string_types)

        s = svg.drawing((100, 100), [], style=self.newstyle)
        self.assertIn(self.newstyle, s)

    def testGroup(self):
        g = svg.group()
        self.assertIsInstance(g, six.string_types)

        g = svg.group(transform="translate(10,10)")
        self.assertIn('transform="translate(10,10)"', g)

        g = svg.group(transform="scale(10)")
        self.assertIn('transform="scale(10)"', g)

    def testAttribs(self):
        args = {'transform': 'translate(10, 10)', 'fill': 'black'}
        self.assertIn('transform="translate(10, 10)"', svg.toattribs(**args))
        self.assertIn('fill="black"', svg.toattribs(**args))

    def testDrawCircle(self):
        point = svg.circle((0.0, 0.0), r=2)
        self.assertIsInstance(point, six.string_types)
        assert 'r="2"' in point
        self.assertIn('cy="0.0"', point)
        assert 'cx="0.0"' in point

    def testDrawPath(self):
        coordinates = [
            (0.0, 0.0),
            (10.0, 0.0),
            (10.0, 10.0),
            (0.0, 10.0),
            (0.0, 0.0),
            'M',
            (4.0, 4.0),
            (4.0, 5.0),
            (5.0, 5.0),
            (5.0, 4.0),
            (4.0, 4.0),
            'Z',
        ]

        path = svg.path(coordinates)
        self.assertIsInstance(path, six.string_types)
        self.assertIn(' M ', path)
        self.assertIn('Z"', path)
        self.assertIn('10.0,0.0', path)

    def assertPartsIn(self, fixture, test):
        for f in fixture:
            self.assertIn(f, test)

    def testRect(self):
        rect = svg.rect((0, 0), 10, 10)
        self.assertIsInstance(rect, type(''))
        fix = '<rect', 'x="0"', 'y="0"', 'width="10"', 'height="10"', '/>'
        self.assertPartsIn(fix, rect)

    def testLine(self):
        line = svg.line((0, 0), (10, 10))
        assert isinstance(line, type(''))
        fix = '<line', 'x1="0"', 'y1="0"', 'x2="10"', 'y2="10"', '/>'
        self.assertPartsIn(fix, line)

    def testText(self):
        text = svg.text('hi', (2, 2))
        assert isinstance(text, type(''))
        fix = '<text', 'x="2"', 'y="2"', 'hi', '</text>'

        self.assertPartsIn(fix, text)

        text = svg.text('hi', (2.22222, 2.222222), precision=2)
        assert isinstance(text, type(''))
        fix = '<text', 'x="2.22"', 'y="2.22"', 'hi', '</text>'

        self.assertPartsIn(fix, text)


if __name__ == '__main__':
    unittest.main()



'''Utilities for svgis'''
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>
from itertools import groupby
from math import ceil, floor

# WGS 84
DEFAULT_GEOID = 4326


def isinf(x):
    """Check if input is infinite."""
    inf = float('inf')
    return x in (inf, inf * -1)


def between(a, b, count=None):
    """Yield <count> points between two floats"""
    jump = (b - a) / float(count or 10.0)

    while a < b:
        yield a
        a += jump


def frange(a, b, step, cover=None):
    """like range, but for floats."""
    while a < b:
        yield a
        a += step

    if cover:
        yield a


def modfloor(inp, mod):
    """A floor function that snaps to steps of size mod."""
    i = inp - (inp % mod)
    return int(floor(i))


def modceil(inp, mod):
    """A ceil function that snaps to steps of size mod."""
    i = inp + mod - (inp % mod)
    return int(ceil(i))


def rnd(i, precision=None):
    """Optionally round to a given precision."""
    if precision is None:
        return i

    return round(i, precision)


def dedupe(array):
    """Use itertools.groupby to remove duplicates in a list."""
    try:
        array = array.tolist()
    except AttributeError:
        pass

    for g in groupby(array):
        yield g[0]


def signed_area(coords):
    """Return the signed area enclosed by a ring using the linear time
    algorithm at http://www.cgafaq.info/wiki/Polygon_Area. A value >= 0
    indicates a counter-clockwise oriented ring."""
    try:
        xs, ys = tuple(map(list, zip(*coords)))
    except ValueError:
        # Attempt to handle a z-dimension
        xs, ys, _ = tuple(map(list, zip(*coords)))

    xs.append(xs[1])
    ys.append(ys[1])
    return sum(xs[i] * (ys[i + 1] - ys[i - 1]) for i in range(1, len(coords))) / 2.0


def clockwise(coords):
    """Check if coordinates move in a clockwise direction."""
    return signed_area(coords) < 0


def counterclockwise(coords):
    """Check if coordinates move in a counterclockwise direction."""
    return signed_area(coords) >= 0



# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>
import logging
import unittest
from os import path
from xml.dom import minidom

from svgis import svgis

EPSG3528 = {'init': 'epsg:3528', 'no_defs': True}
# logging.getLogger('svgis').setLevel(logging.DEBUG)


class ProjectionDrawTestCase(unittest.TestCase):

    bounds = {
        # IL E (ft)
        2790: (347026, 556571, 364500, 592793),
        # NAD 83
        4269: (-87.8, 41.7, -87.5, 41.9),
    }

    def setUp(self):
        self.files = [
            path.join(path.dirname(__file__), 'fixtures', 'chicago_bounds_2790.json'),
            path.join(path.dirname(__file__), 'fixtures', 'cook_bounds_4269.json'),
        ]

        svgis.STYLE = ''

    def testDrawWithReProjection(self):
        s = svgis.SVGIS(self.files, self.bounds[2790], out_crs=EPSG3528, scalar=100)
        svg = s.compose()

        self.assertIn('points', svg)
        i = svg.index('points')
        self.assertIn('points', svg[i:])

        polygons = minidom.parseString(svg).getElementsByTagName('polygon')

        self.assertEqual(len(polygons), 2)

        self.assertIn('points', dict(polygons[0].attributes.items()))
        self.assertIn('points', dict(polygons[1].attributes.items()))

    def testDrawWithReProjectionRepeat(self):
        s = svgis.SVGIS(self.files[1], self.bounds[4269], out_crs=EPSG3528, scalar=100)
        s.compose()
        u = s.compose()
        self.assertIn('points', u)
        i = u.index('points')
        self.assertIn('points', u[i:])

    def testDrawWithSameProjection(self):
        s = svgis.SVGIS(self.files, crs=2790)
        a = s.compose()
        self.assertIn('points', a)
        i = a.index('points')
        self.assertIn('points', a[i:])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests on the svgis command line tool"""
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>
# pylint: disable=duplicate-code
import io
import os
import re
import sys
import unittest
from xml.dom import minidom

import click.testing
import fiona.errors
from click import BadParameter

import svgis.cli

PROJECTION = '+proj=lcc +lat_1=20 +lat_2=60 +lat_0=40 +lon_0=-96 +x_0=0 +y_0=0 +datum=NAD83 +units=m +no_defs'
BOUNDS = (-124.0, 20.5, -64.0, 49.0)


class CliTestCase(unittest.TestCase):
    runner = click.testing.CliRunner()
    fixture = 'tests/fixtures/test.svg'

    shp = 'tests/fixtures/cb_2014_us_nation_20m.json'
    dc = 'tests/fixtures/tl_2015_11_place.json'
    css = 'polygon{fill:green}'

    def setUp(self):
        self.assertTrue(os.path.exists(self.dc))
        self.assertTrue(os.path.exists(self.fixture))
        self.assertTrue(os.path.exists(self.shp))

    def invoke(self, argument, **kwargs):
        return self.runner.invoke(svgis.cli.main, argument, catch_exceptions=False, **kwargs)

    def testSvgStyle(self):
        sys.stdout = io.StringIO()
        self.invoke(['style', '--style', self.css, self.fixture, 'tmp.svg'])
        try:
            with open('tmp.svg') as f:
                self.assertIn(self.css, f.read()[0:1000])
        finally:
            os.remove('tmp.svg')

    def testSvgScale(self):
        self.invoke(['scale', '--scale', '123', self.fixture, 'tmp.svg'])
        try:
            with open('tmp.svg') as f:
                self.assertIn('scale(123)', f.read()[0:1000])

        finally:
            os.remove('tmp.svg')

    def testBounds(self):
        result = self.invoke(['bounds', self.shp])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output.strip(), '-179.174265 17.913769 179.773922 71.352561')

    def testSvgProjectUtm(self):
        p = self.invoke(['project', '--method', 'utm', '--', '-110.277906', '35.450777', '-110.000477', '35.649030'])
        expected = set('+proj=utm +zone=12 +datum=WGS84 +units=m +no_defs +type=crs'.split(' '))
        self.assertSetEqual(set(p.output.strip().split(' ')), expected)

    def testSvgProject(self):
        p = self.invoke(['project', '--', '-110.277906', '35.450777', '-110.000477', '35.649030'])

        self.assertEqual(p.exit_code, 0)

        expected = set(
            [
                '+type=crs',
                '+lat_0=35.64903',
                '+x_0=0',
                '+units=m',
                '+lon_0=-111',
                '+towgs84=0,0,0,0,0,0,0',
                '+y_0=0',
                '+lat_1=35.64903',
                '+proj=lcc',
                '+no_defs',
                '+lat_2=35.450777',
                '+ellps=GRS80',
            ]
        )
        self.assertSetEqual(set(p.output.strip().split(' ')), expected)

    def testCliDraw(self):
        self.invoke(
            ['draw', '--crs', PROJECTION, '--scale', '1000', self.shp, '-o', 'tmp.svg', '--viewbox', '--bounds']
            + [str(b) for b in BOUNDS]
        )

        try:
            result = minidom.parse('tmp.svg').getElementsByTagName('svg').item(0)

            fixture = minidom.parse(self.fixture).getElementsByTagName('svg').item(0)

            result_vb = [float(x) for x in result.attributes.get('viewBox').value.split(',')]
            fixture_vb = [float(x) for x in fixture.attributes.get('viewBox').value.split(',')]

            for r, f in zip(result_vb, fixture_vb):
                self.assertAlmostEqual(r, f, 5, 'viewbox doesnt match fixture')

        finally:
            os.remove('tmp.svg')

    def testDrawProjected(self):
        f = os.path.expanduser('~/tmp.svg')
        result = self.invoke(['draw', self.dc, '--output', f, '--precision', '10'])

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(os.path.exists(f))

        try:
            with open(f, encoding="utf8") as g:
                svg = g.read()
                match = re.search(r'points="([^"]+)"', svg)
                self.assertIsNotNone(match)

            result = match.groups()[0]
            points = [[float(x) for x in p.split(',')] for p in result.split(' ')]

            with fiona.open(self.dc) as src:
                ring = next(iter(src))['geometry']['coordinates'][0]

            for points in zip(ring, points):
                for z in zip(*points):
                    self.assertAlmostEqual(*z)

        finally:
            os.remove(f)

    def testCliHelp(self):
        result = self.invoke(('--help',))
        self.assertEqual(result.exit_code, 0)

        result = self.invoke(['style', '--help'])
        assert result.exit_code == 0

        result = self.invoke(['draw', '--help'])
        assert result.exit_code == 0

        result = self.invoke(['project', '--help'])
        assert result.exit_code == 0

        result = self.invoke(['scale', '--help'])
        assert result.exit_code == 0

    def testErrs(self):
        with self.assertRaises(fiona.errors.DriverError):
            self.invoke(('draw', 'lksdjlksjdf'))

        with self.assertRaises(fiona.errors.DriverError):
            self.invoke(('draw', 'zip://lksdjlksjdf.zip/foo'))

    def test_validate_posint(self):
        with self.assertRaises(BadParameter):
            svgis.cli.validate_posint(None, None, -1)

        with self.assertRaises(BadParameter):
            svgis.cli.validate_posint(None, None, 0.5)

        self.assertEqual(svgis.cli.validate_posint(None, None, 1), 1)


if __name__ == '__main__':
    unittest.main()



# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>
# pylint: disable=unused-import
import functools
import unittest

from svgis import transform

try:
    import shapely.geometry

    NO_SHAPELY = False

except ImportError:
    NO_SHAPELY = True
try:
    import visvalingamwyatt

    VW = True
except ImportError:
    VW = False


class ClipTestCase(unittest.TestCase):
    """Test svgis.transform.clip"""

    def setUp(self):
        self.bounds = (1, 1, 9, 9)
        self.coords = [[(2, 2), (100, 2), (11, 11), (12, 12), (2, 10), (2, 2)]]
        self.fixture = shapely.geometry.Polygon(((2.0, 9.0), (9.0, 9.0), (9.0, 2.0), (2.0, 2.0), (2.0, 9.0)))
        self.gen = (x for x in self.coords[0])

    @unittest.skipIf(NO_SHAPELY, "Shapely not installed")
    def testShapely(self):
        minx, miny, maxx, maxy = self.bounds
        bounds = {
            "type": "Polygon",
            "coordinates": [[(minx, miny), (minx, maxy), (maxx, maxy), (maxx, miny), (minx, miny)]],
        }
        coords = {"type": "Polygon", "coordinates": self.coords}

        b = shapely.geometry.shape(bounds)
        c = shapely.geometry.shape(coords)
        result = b.intersection(c)
        self.assertTrue(result.equals(self.fixture), f"{result} == {self.fixture}")

    @unittest.skipIf(NO_SHAPELY, "Shapely not installed")
    def testClip(self):
        clipped = transform.clip({"type": "Polygon", "coordinates": self.coords}, self.bounds)
        result = shapely.geometry.Polygon(clipped['coordinates'][0])
        self.assertTrue(result.equals(self.fixture), f"{result} == {self.fixture}")


class SimplifyTestCase(unittest.TestCase):
    """Test svgis.transform.simplifier"""

    def testSimplifyNone(self):
        a = transform.simplifier(None)
        self.assertIsNone(a)

    @unittest.skipIf(not VW, "visvalingamwyatt is not installed")
    def testSimplifyTypeVV(self):
        c = transform.simplifier(50)
        self.assertIsInstance(c, functools.partial)

    @unittest.skipIf(VW, "visvalingamwyatt is installed")
    def testSimplifyTypeNoVW(self):
        c = transform.simplifier(50)
        self.assertIsNone(c)


if __name__ == '__main__':
    unittest.main()



# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>
import logging
import unittest

from svgis import SVGIS, draw, errors


class ErrorTestCase(unittest.TestCase):

    feature = {'geometry': {'type': 'Bizarro', 'coordinates': [[(1, 2), (3, 4)], [(7, 8), (9, 10)]]}, 'properties': {}}

    def setUp(self):
        logging.getLogger('svgis').setLevel(logging.CRITICAL)

    def testDrawInvalidGeometry(self):
        with self.assertRaises(errors.SvgisError):
            draw.geometry(self.feature['geometry'])

    def testSvgisDrawInvalidGeometry(self):
        a = SVGIS([]).feature(self.feature, [], [])
        assert a == u''


if __name__ == '__main__':
    unittest.main()



"""SVGIS error type(s)."""
# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>


class SvgisError(Exception):
    """The SVGIS error type currently does nothing special."""



# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>
import unittest

import tinycss2

try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree

from svgis import dom

from . import TEST_CSS, TEST_SVG


class DomTestCase(unittest.TestCase):

    svg = TEST_SVG
    css = TEST_CSS

    def setUp(self):
        self.rules = tinycss2.parse_stylesheet(self.css, skip_whitespace=True, skip_comments=True)
        self.document = etree.fromstring(self.svg)

    def testSerializeToken(self):
        rules = tinycss2.parse_stylesheet("polygon{fill:orange}")
        token = rules[0].prelude[0]
        assert token.type == 'ident'

        assert ('svg|' + token.serialize()) == 'svg|polygon'

        prelude_value = dom.serialize_token(token)
        self.assertEqual(prelude_value, 'svg|polygon')

        prelude = dom.serialize_prelude(rules[0])
        self.assertEqual(prelude, 'svg|polygon')

        pretoken = tinycss2.ast.LiteralToken(1, 1, '.')
        serialized = dom.serialize_token(rules[0].prelude[0], pretoken)
        self.assertEqual(serialized, 'polygon')

        pretoken = tinycss2.ast.LiteralToken(1, 1, '/')
        serialized = dom.serialize_token(rules[0].prelude[0], pretoken)
        self.assertEqual(serialized, 'polygon')

    def testApplyRule(self):
        rules = tinycss2.parse_stylesheet("polygon {fill: orange}", skip_whitespace=True, skip_comments=True)
        svg = etree.fromstring(
            '<svg baseProfile="full" height="1" version="1.1" xmlns="http://www.w3.org/2000/svg">'
            '<polygon class="test" points="3,2 -2,6 8,-1 8,2 4,1 3,2" />'
            '</svg>'
        )
        dom.apply_rules(svg, rules)
        text = etree.tostring(svg).decode('ascii')
        self.assertIn('orange', text)

    def testApplyIdentRule(self):
        dom.apply_rules(self.document, self.rules)
        polygon = self.document.xpath('.//svg:polygon', namespaces={'svg': 'http://www.w3.org/2000/svg'})[0]

        self.assertIsNotNone(polygon)

        self.assertIn('fill:orange', polygon.attrib['style'])

    def testApplyHashRule(self):
        dom.apply_rules(self.document, self.rules)
        polyline = self.document.xpath(".//*[@id='meow']")[0]
        self.assertIsNotNone(polyline)
        self.assertIn('stroke-opacity:0.50', polyline.attrib.get('style', ''))

    def testChildToken(self):
        css = "#foo > #baz { stroke: green }"
        rules = tinycss2.parse_stylesheet(css, skip_whitespace=True, skip_comments=True)
        dom.apply_rules(self.document, rules)
        polyline = self.document.find(".//*[@id='baz']")
        self.assertIn('stroke:green', polyline.attrib.get('style', ''))

    def testApplyRadiusRule(self):
        css = tinycss2.parse_stylesheet('circle { r: 1 }')
        dom.apply_rules(self.document, css)
        circ = self.document.find('.//circle', namespaces=self.document.nsmap)
        self.assertIsNotNone(circ)
        self.assertIn('1', circ.attrib.get('r', ''))


if __name__ == '__main__':
    unittest.main()
